﻿/*
 *
 *
 *    Simulator.cs
 *
 *    Provides test classes for the application.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Threading;
using System.Diagnostics;

using KK5JY.IO;

namespace KK5JY.AutoCap {
	/// <summary>
	/// An empty radio.
	/// </summary>
	public class NullRadio : IRadio, IReportsVersion {
		/// <summary>
		/// Reads the frequency from the radio (in this case, always zero).
		/// </summary>
		public decimal Frequency { get { return 0.0m; } }

		/// <summary>
		/// The version information.
		/// </summary>
		public string VersionString { get { return "None"; } }
	}

	/// <summary>
	/// A test harness that simulates a loop.
	/// </summary>
	public class TestHardware : ISensor, IMotor, IStepperMotor, IRadio, IReportsVersion, IStart, IDisposable {
		private static TestHardware s_Instance;
		public static TestHardware Instance {
			get {
				if (s_Instance == null) {
					s_Instance = new TestHardware();
				}
				return s_Instance;
			}
		}

		/// <summary>
		/// Raised for each new SWR measurement.
		/// </summary>
		public event EventHandler<SensorDataEvent> SensorData;

		/// <summary>
		/// Stepper rate.
		/// </summary>
		public const decimal StepsPerRevolution = 100m;

		/// <summary>
		/// Antenna movement rate.
		/// </summary>
		public const decimal HzPerRevolution = 50000m;

		/// <summary>
		/// How fast the motor can move the antenna frequency per simulation cycle.
		/// </summary>
		public decimal HzPerCycle { get; set; }

		/// <summary>
		/// The current antenna center frequency.
		/// </summary>
		public decimal AntennaFrequency { get; set; }

		/// <summary>
		/// The frequency of the antenna when the motor position is zero.
		/// </summary>
		public decimal AntennaCenterFrequency { get; set; }

		/// <summary>
		/// The current antenna bandwidth.
		/// </summary>
		public decimal AntennaBandwidth { get; set; }

		/// <summary>
		/// Determines the SWR edges of the bandwidth curve.
		/// That is, the 'AntennaBandwidth' is the width of
		/// the curve with 'AntennaEdgeSWR' on both edges,
		/// and 1.0:1 in the middle.
		/// </summary>
		public decimal AntennaEdgeSWR { get; set; }

		/// <summary>
		/// The current carrier frequency.
		/// </summary>
		public decimal CarrierFrequency { get; set; }

		/// <summary>
		/// The current carrier power.
		/// </summary>
		public decimal CarrierPower { get; set; }

		/// <summary>
		/// The current SWR.
		/// </summary>
		public decimal SWR { get; private set; }

		/// <summary>
		/// The 'radio frequency'.
		/// </summary>
		public decimal Frequency {
			get { return CarrierFrequency; }
		}

		/// <summary>
		/// Enables automatic SWR polling.
		/// </summary>
		public bool AutoPoll {
			get { return m_AutoPoll; }
			set {
				if (value) {
					Start();
				} else {
					Stop();
				}
			}
		}

		/// <summary>
		/// Indicates that the sensor should only try to read SWR (not power, reactance, etc.)
		/// </summary>
		public bool ReadOnlySWR { get; set; }

		/// <summary>
		/// Reverse the motor effort for movement commands.
		/// </summary>
		public bool MotorReversed { get; set; }

		/// <summary>
		/// Get the stepper position.
		/// </summary>
		public long Position {
            get {
                lock (Instance)
                    return Convert.ToInt64(mPosition);
            }
            private set {
                lock (Instance)
                    mPosition = Convert.ToDouble(value);
            }
		}

		/// <summary>
		/// The version string.
		/// </summary>
		public string VersionString {
			get { return "AutoCap 1.0 Simulator"; }
		}

		public event EventHandler<PositionEventArgs> PositionUpdate;

		#region Private Fields
		// the worker thread
		private Thread m_Thread;

		// the current motor speed
		private decimal m_MotorSpeed;

		// backing var for AutoPoll
		private bool m_AutoPoll;

		// the motor position
		private double mPosition;
		#endregion

		#region Public Methods
		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="f0">Initial antenna frequency.</param>
		/// <param name="bw">Initial antenna bandwidth.</param>
		private TestHardware(decimal f0 = 10000000.0m, decimal bw = 25000.0m) {
			AntennaCenterFrequency = f0; // Hz
			AntennaFrequency = f0;  // Hz
			AntennaBandwidth = bw;  // Hz
			AntennaEdgeSWR = 10.0m; // :1
			HzPerCycle = 500m;
			CarrierPower = 10m;  // W
			CarrierFrequency = f0 + 500m; // just a little off
			mPosition = 0;
		}

		/// <summary>
		/// Dispose the object.
		/// </summary>
		public void Dispose() {
			Stop();
		}

		/// <summary>
		/// Start the simulator.
		/// </summary>
		public void Start() {
			m_AutoPoll = true;
			if (m_Thread != null) return;
			m_Thread = new Thread(Worker);
			m_Thread.Name = "KK5JY.AutoCap.TestHardware.Worker";
			m_Thread.Start();
		}

		/// <summary>
		/// Stop the simulator.
		/// </summary>
		public void Stop() {
			m_AutoPoll = false;
			if (m_Thread == null) return;
			m_Thread.Join();
			m_Thread = null;
		}

		/// <summary>
		/// Poll for a new SWR reading.
		/// </summary>
		public void Poll() {
			CalculateSWR();
			if (SensorData != null) {
				var update = new SensorDataEvent {
					Data = new SensorData {
						SWR = SWR,
						ID = "TEST",
					}
				};
				if (!ReadOnlySWR) {
					update.Data.Power = CarrierPower;
				}
				SensorData(this, update);
			}
		}

		/// <summary>
		/// Move the motor at the given speed.
		/// </summary>
		/// <param name="speed">The speed as a signed scalar.</param>
		public bool Move(decimal speed = 0m) {
			if (MotorReversed)
				speed *= -1m;
			m_MotorSpeed = speed;
			return true;
		}

		/// <summary>
		/// Step the motor at the given speed.
		/// </summary>
		/// <param name="speed">The speed as a signed scalar.</param>
		/// <param name="steps">The number of steps.</param>
		/// <returns></returns>
        public bool Step(decimal speed, ulong steps)
        {
            if (MotorReversed)
                speed *= -1;
            lock (Instance)
                mPosition += Convert.ToDouble(Math.Sign(speed) * Convert.ToInt64(steps));
            UpdateFrequencyFromPosition();
            return true;
        }

		/// <summary>
		/// Reset the odometer.
		/// </summary>
		public bool ResetPosition() {
            lock (Instance)
                mPosition = 0;
			UpdateFrequencyFromPosition();
			return true;
		}

		/// <summary>
		/// Reset the odometer.
		/// </summary>
		public bool Seek(long target) {
            lock (Instance)
                mPosition = Convert.ToDouble(target);
			UpdateFrequencyFromPosition();
			return true;
		}

		/// <summary>
		/// Request the current position.
		/// </summary>
		public bool QueryPosition() {
			if (PositionUpdate != null) {
				PositionUpdate(this, new PositionEventArgs(Position));
			}
			return true;
		}
		#endregion

		#region Utilities
		/// <summary>
		/// Calculate the SWR value for the given environment.
		/// </summary>
		private void CalculateSWR() {
			if (CarrierPower == 0m) {
				SWR = 1.0m;
				return;
			}

			// edge checking
			var deltaF = (CarrierFrequency - AntennaFrequency) / AntennaBandwidth;
			if (deltaF > 0.5m) deltaF = 0.5m;
			if (deltaF < -0.5m) deltaF = -0.5m;

			// do the actual calculation
			SWR = Convert.ToDecimal(
				NormalizedSWR(
					Convert.ToDouble(deltaF),
					Convert.ToDouble(AntennaEdgeSWR)
			));
		}

		/// <summary>
		/// Calculates a normalized SWR value for a given offset.
		/// </summary>
		/// <remarks>
		/// The function calculates an SWR value for a given offset
		/// which is a normalized fraction of the 2:1 SWR bandwidth.
		/// This is a 'raised cosine' approximation of the SWR
		/// response of an antenna within its 2:1 bandwidth.
		/// </remarks>
		/// <param name="offset">The offset from the center frequency.</param>
		/// <param name="mult">The multiplier to apply, to approximate
		/// the bandwidth measured at a higher SWR limit.</param>
		private double NormalizedSWR(double offset, double edge = 2.0) {
			return 1.0 + ((edge - 1.0) * (1.0 - Math.Cos(Math.PI * offset)));
		}

		/// <summary>
		/// Read the current position and set the antenna frequency accordingly.
		/// </summary>
        private void UpdateFrequencyFromPosition() {
            lock (Instance)
                AntennaFrequency = AntennaCenterFrequency + ((Convert.ToDecimal(mPosition) / StepsPerRevolution) * HzPerRevolution);
        }
		#endregion

		#region Threads
		/// <summary>
		/// The worker thread.
		/// </summary>
		private void Worker() {
			while (m_AutoPoll) {
				// recalculate the starting SWR
				Poll();

				// move the motor
				if (m_MotorSpeed != 0m) {
					// not very precise, but sufficiently realistic for now
                    lock (Instance)
					    mPosition += Convert.ToDouble(m_MotorSpeed * StepsPerRevolution * HzPerCycle / HzPerRevolution);
					UpdateFrequencyFromPosition();
					Debug.WriteLine("DEBUG: Antenna Frequency: {0}", AntennaFrequency);
				}

				// and sleep briefly
				Thread.Sleep(50);
			}
		}
		#endregion
	}
}
